@@ -43,6 +43,7 @@ __all__ = ["reimport", "modified"]
 import sys
 import os
 import gc
+import imp
 import inspect
 import weakref
 import traceback
@@ -50,7 +51,7 @@ import time
 
 
 
-__version__ = "1.2"
+__version__ = "1.3"
 __author__ = "Peter Shinners <pete@shinners.org>"
 __license__ = "MIT"
 __url__ = "http://code.google.com/p/reimport"
@@ -82,7 +83,7 @@ def reimport(*modules):
         If a package module defines __reimported__ it must be a callable
         function that accepts one argument and returns a bool. The argument
         is the reference to the old version of that module before any
-        cleanup has happend. The function should normally return True to
+        cleanup has happened. The function should normally return True to
         allow the standard reimport cleanup. If the function returns false
         then cleanup will be disabled for only that module. Any exceptions
         raised during the callback will be handled by traceback.print_exc,
@@ -126,72 +127,112 @@ def reimport(*modules):
         
         compile(data, pyname, "exec", 0, False)  # Let this raise exceptions
 
-    # Move modules out of sys
-    oldModules = {}
-    for name in reloadNames:
-        oldModules[name] = sys.modules.pop(name)
-    ignores = (id(oldModules), id(__builtins__))
-    prevNames = set(sys.modules)
-
-    # Reimport modules, trying to rollback on exceptions
+    # Begin changing things. We "grab the GIL", so other threads
+    # don't get a chance to see our half-baked universe
+    imp.acquire_lock()
+    prevInterval = sys.getcheckinterval()
+    sys.setcheckinterval(min(sys.maxint, 0x7fffffff))
     try:
-        for name in reloadNames:
-            if name not in sys.modules:
-                __import__(name)
 
-    except StandardError:
-        # Try to dissolve any newly import modules and revive the old ones
+        # Python will munge the parent package on import. Remember original value
+        parentValues = []
+        parentPackageDeleted = lambda: None
+        for name in reloadNames:
+            parentPackageName = name.rsplit(".", 1)
+            if len(parentPackageName) == 2:
+                parentPackage = sys.modules.get(parentPackageName[0], None)
+                parentValue = getattr(parentPackage, parentPackageName[1], parentPackageDeleted)
+                if parentValue != sys.modules[name]:
+                    parentValues.append((parentPackage, parentPackageName[1], parentValue))
+                parentPackage = parentValue = None
+
+        # Move modules out of sys
+        oldModules = {}
+        for name in reloadNames:
+            oldModules[name] = sys.modules.pop(name)
+        ignores = (id(oldModules),)
+        prevNames = set(sys.modules)
+
+        # Reimport modules, trying to rollback on exceptions
+        try:
+            try:
+                for name in reloadNames:
+                    if name not in sys.modules:
+                        __import__(name)
+
+            except StandardError:
+                # Try to dissolve any newly import modules and revive the old ones
+                newNames = set(sys.modules) - prevNames
+                newNames = _package_depth_sort(newNames, True)
+                for name in newNames:
+                    backoutModule = sys.modules.pop(name, None)
+                    if backoutModule is not None:
+                        _unimport(backoutModule, ignores)
+                    del backoutModule
+
+                sys.modules.update(oldModules)
+                raise
+
+        finally:
+            # Fix Python automatically shoving of children into parent packages
+            for parentPackage, name, value in parentValues:
+                if value == parentPackageDeleted:
+                    try:
+                        delattr(parentPackage, name)
+                    except AttributeError:
+                        pass
+                else:
+                    setattr(parentPackage, name, value)
+            parentValues = parentPackage = parentPackageDeleted = value = None
+
         newNames = set(sys.modules) - prevNames
         newNames = _package_depth_sort(newNames, True)
+
+        # Update timestamps for loaded time
+        now = time.time() - 1.0
         for name in newNames:
-            _unimport_module(sys.modules[name], ignores)
-            assert name not in sys.modules
+            _module_timestamps[name] = (now, True)
 
-        sys.modules.update(oldModules)
-        raise
+        # Push exported namespaces into parent packages
+        pushSymbols = {}
+        for name in newNames:
+            oldModule = oldModules.get(name)
+            if not oldModule:
+                continue
+            parents = _find_parent_importers(name, oldModule, newNames)
+            pushSymbols[name] = parents
+        for name, parents in pushSymbols.iteritems():
+            for parent in parents:
+                oldModule = oldModules[name]
+                newModule = sys.modules[name]
+                _push_imported_symbols(newModule, oldModule, parent)
 
-    newNames = set(sys.modules) - prevNames
-    newNames = _package_depth_sort(newNames, True)
+        # Rejigger the universe
+        for name in newNames:
+            old = oldModules.get(name)
+            if not old:
+                continue
+            new = sys.modules[name]
+            rejigger = True
+            reimported = getattr(new, "__reimported__", None)
+            if reimported:
+                try:
+                    rejigger = reimported(old)
+                except StandardError:
+                    # What else can we do? the callbacks must go on
+                    # Note, this is same as __del__ behaviour. /shrug
+                    traceback.print_exc()
 
-    # Update timestamps for loaded time
-    now = time.time() - 1.0
-    for name in newNames:
-        _module_timestamps[name] = (now, True)
+            if rejigger:
+                _rejigger_module(old, new, ignores)
+            else:
+                _unimport_module(new, ignores)
 
-    # Push exported namespaces into parent packages
-    pushSymbols = {}
-    for name in newNames:
-        oldModule = oldModules.get(name)
-        if not oldModule:
-            continue
-        parents = _find_parent_importers(name, oldModule, newNames)
-        pushSymbols[name] = parents
-    for name, parents in pushSymbols.iteritems():
-        for parent in parents:
-            oldModule = oldModules[name]
-            newModule = sys.modules[name]
-            _push_imported_symbols(newModule, oldModule, parent)
-
-    # Rejigger the universe
-    for name in newNames:
-        old = oldModules.get(name)
-        if not old:
-            continue
-        new = sys.modules[name]
-        rejigger = True
-        reimported = getattr(new, "__reimported__", None)
-        if reimported:
-            try:
-                rejigger = reimported(old)
-            except StandardError:
-                # What else can we do? the callbacks must go on
-                # Note, this is same as __del__ behaviour. /shrug
-                traceback.print_exc()
-
-        if rejigger:
-            _rejigger_module(old, new, ignores)
-        else:
-            _unimport_module(new, ignores)
+    finally:
+        # Restore the GIL
+        imp.release_lock()
+        sys.setcheckinterval(prevInterval)
+        time.sleep(0)
 
 
 
@@ -241,6 +282,12 @@ def modified(path=None):
     return modules
 
 
+def _safevars(obj):
+    try:
+        return vars(obj)
+    except TypeError:
+        return {}
+
 
 def _is_code_module(module):
     """Determine if a module comes from python code"""
@@ -302,10 +349,10 @@ def _package_depth_sort(names, reverse):
 
 
 def _find_module_exports(module):
-    all = getattr(module, "__all__", ())
-    if not all:
-        all = [n for n in dir(module) if n[0] != "_"]
-    return set(all)
+    allNames = getattr(module, "__all__", ())
+    if not allNames:
+        allNames = [n for n in dir(module) if n[0] != "_"]
+    return set(allNames)
 
 
 
@@ -334,6 +381,7 @@ def _find_parent_importers(name, oldModule, newNames):
     return parents
 
 
+
 def _push_imported_symbols(newModule, oldModule, parent):
     """Transfer changes symbols from a child module to a parent package"""
     # This assumes everything in oldModule is already found in parent
@@ -344,15 +392,26 @@ def _push_imported_symbols(newModule, oldModule, parent):
     for name in oldExports - newExports:
         delattr(parent, name)
     
+    # Find symbols in new module, use placeholder if missing
+    symbols = {}
+    for name in newExports:
+        try:
+            symbols[name] = getattr(newModule, name)
+        except AttributeError:
+            holder = type(name, (_MissingAllReference,),
+                        {"__module__":newModule.__name__})
+            symbols[name] = holder()
+        
+    
     # Add new symbols
     for name in newExports - oldExports:
-        setattr(parent, name, getattr(newModule, name))
+        setattr(parent, name, symbols[name])
     
     # Update existing symbols
     for name in newExports & oldExports:
         oldValue = getattr(oldModule, name)
         if getattr(parent, name) is oldValue:
-            setattr(parent, name, getattr(newModule, name))
+            setattr(parent, name, symbols[name])
     
 
 
@@ -363,14 +422,13 @@ def _push_imported_symbols(newModule, oldModule, parent):
 def _rejigger_module(old, new, ignores):
     """Mighty morphin power modules"""
     __internal_swaprefs_ignore__ = "rejigger_module"
-    oldVars = vars(old)
-    newVars = vars(new)
+    oldVars = _safevars(old)
+    newVars = _safevars(new)
     ignores += (id(oldVars),)
     old.__doc__ = new.__doc__
 
     # Get filename used by python code
     filename = new.__file__
-    fileext = os.path.splitext(filename)
 
     for name, value in newVars.iteritems():
         if name in oldVars:
@@ -380,16 +438,17 @@ def _rejigger_module(old, new, ignores):
 
             if _from_file(filename, value):
                 if inspect.isclass(value):
-                    _rejigger_class(oldValue, value, ignores)
+                    if inspect.isclass(oldValue):
+                        _rejigger_class(oldValue, value, ignores)
                     
                 elif inspect.isfunction(value):
-                    _rejigger_func(oldValue, value, ignores)
+                    if inspect.isfunction(oldValue):
+                        _rejigger_func(oldValue, value, ignores)
         
         setattr(old, name, value)
 
-    for name in oldVars.keys():
+    for name, value in oldVars.items():
         if name not in newVars:
-            value = getattr(old, name)
             delattr(old, name)
             if _from_file(filename, value):
                 if inspect.isclass(value) or inspect.isfunction(value):
@@ -412,12 +471,19 @@ def _from_file(filename, value):
 def _rejigger_class(old, new, ignores):
     """Mighty morphin power classes"""
     __internal_swaprefs_ignore__ = "rejigger_class"    
-    oldVars = vars(old)
-    newVars = vars(new)
+    oldVars = _safevars(old)
+    newVars = _safevars(new)
     ignores += (id(oldVars),)    
 
+    slotted = hasattr(old, "__slots__") and isinstance(old.__slots__, tuple)
+    ignoreAttrs = ["__dict__", "__doc__", "__weakref__"]
+    if slotted:
+        ignoreAttrs.extend(old.__slots__)
+        ignoreAttrs.append("__slots__")
+    ignoreAttrs = tuple(ignoreAttrs)
+
     for name, value in newVars.iteritems():
-        if name in ("__dict__", "__doc__", "__weakref__"):
+        if name in ignoreAttrs:
             continue
 
         if name in oldVars:
@@ -433,9 +499,8 @@ def _rejigger_class(old, new, ignores):
 
         setattr(old, name, value)
     
-    for name in oldVars.keys():
+    for name, value in oldVars.items():
         if name not in newVars:
-            value = getattr(old, name)
             delattr(old, name)
             _remove_refs(value, ignores)
 
@@ -454,10 +519,21 @@ def _rejigger_func(old, new, ignores):
 
 
 
+def _unimport(old, ignores):
+    """Unimport something, mainly used to rollback a reimport"""
+    if isinstance(old, type(sys)):
+        _unimport_module(old, ignores)
+    elif inspect.isclass(old):
+        _unimport_class(old, ignores)
+    else:
+        _remove_refs(old, ignores)
+    
+
+
 def _unimport_module(old, ignores):
     """Remove traces of a module"""
     __internal_swaprefs_ignore__ = "unimport_module"
-    oldValues = vars(old).values()
+    oldValues = _safevars(old).values()
     ignores += (id(oldValues),)    
 
     # Get filename used by python code
@@ -484,7 +560,7 @@ def _unimport_module(old, ignores):
 def _unimport_class(old, ignores):
     """Remove traces of a class"""
     __internal_swaprefs_ignore__ = "unimport_class"    
-    oldItems = vars(old).items()
+    oldItems = _safevars(old).items()
     ignores += (id(oldItems),)    
 
     for name, value in oldItems:
@@ -501,9 +577,46 @@ def _unimport_class(old, ignores):
 
 
 
+
+class _MissingAllReference(object):
+    """This is a stub placeholder for objects added to __all__ but
+        are not actually found.
+        """
+    def __str__(self, *args):
+        raise AttributeError("%r missing from module %r" %
+                    (type(self).__name__, type(self).__module__))
+    __nonzero__ = __hash__ = __id__ = __cmp__ = __len__ = __iter__ = __str__
+    __repr__ = __int__ = __getattr__ = __setattr__ = __delattr__ = __str__
+    
+
+
 _recursive_tuple_swap = set()
 
 
+def _bonus_containers():
+    """Find additional container types, if they are loaded. Returns
+        (deque, defaultdict).
+        Any of these will be None if not loaded. 
+        """
+    deque = defaultdict = None
+    collections = sys.modules.get("collections", None)
+    if collections:
+        deque = getattr(collections, "collections", None)
+        defaultdict = getattr(collections, "defaultdict", None)
+    return deque, defaultdict
+
+
+
+def _find_sequence_indices(container, value):
+    """Find indices of value in container. The indices will
+        be in reverse order, to allow safe editing.
+        """
+    indices = []
+    for i in range(len(container)-1, -1, -1):
+        if container[i] is value:
+            indices.append(i)
+    return indices
+
 
 def _swap_refs(old, new, ignores):
     """Swap references from one object to another"""
@@ -520,6 +633,8 @@ def _swap_refs(old, new, ignores):
                 _swap_refs(oldRef, newRef, ignores + (id(refs),))
     del refs
 
+    deque, defaultdict = _bonus_containers()
+
     # Swap through garbage collector
     referrers = gc.get_referrers(old)
     for container in referrers:
@@ -527,12 +642,8 @@ def _swap_refs(old, new, ignores):
             continue
         containerType = type(container)
         
-        if containerType is list:
-            while True:
-                try:
-                    index = container.index(old)
-                except ValueError:
-                    break
+        if containerType is list or containerType is deque:
+            for index in _find_sequence_indices(container, old):
                 container[index] = new
         
         elif containerType is tuple:
@@ -543,18 +654,14 @@ def _swap_refs(old, new, ignores):
             _recursive_tuple_swap.add(id(orig))
             try:
                 container = list(container)
-                while True:
-                    try:
-                        index = container.index(old)
-                    except ValueError:
-                        break
+                for index in _find_sequence_indices(container, old):
                     container[index] = new
                 container = tuple(container)
                 _swap_refs(orig, container, ignores + (id(referrers),))
             finally:
                 _recursive_tuple_swap.remove(id(orig))
         
-        elif containerType is dict:
+        elif containerType is dict or containerType is defaultdict:
             if "__internal_swaprefs_ignore__" not in container:
                 try:
                     if old in container:
@@ -568,15 +675,19 @@ def _swap_refs(old, new, ignores):
         elif containerType is set:
             container.remove(old)
             container.add(new)
-            
-        elif containerType == type:
+
+        elif containerType is type:
             if old in container.__bases__:
                 bases = list(container.__bases__)
                 bases[bases.index(old)] = new
                 container.__bases__ = tuple(bases)
         
         elif type(container) is old:
-            container.__class__ = new
+            try:
+                container.__class__ = new
+            except TypeError:
+                # Type error happens on slotted classes
+                pass
         
         elif containerType is _InstanceType:
             if container.__class__ is old:
@@ -589,10 +700,10 @@ def _remove_refs(old, ignores):
     __internal_swaprefs_ignore__ = "remove_refs"
     
     # Ignore builtin immutables that keep no other references
-    _isinst = isinstance
-    if (old is None or _isinst(old, int) or _isinst(old, basestring)
-                or _isinst(old, float) or _isinst(old, complex)):
+    if old is None or isinstance(old, (int, basestring, float, complex)):
         return
+
+    deque, defaultdict = _bonus_containers()
     
     # Remove through garbage collector
     for container in gc.get_referrers(old):
@@ -600,34 +711,27 @@ def _remove_refs(old, ignores):
             continue
         containerType = type(container)
 
-        if containerType == list:
-            while True:
-                try:
-                    container.remove(old)
-                except ValueError:
-                    break
+        if containerType is list or containerType is deque:
+            for index in _find_sequence_indices(container, old):
+                del container[index]
         
-        elif containerType == tuple:
+        elif containerType is tuple:
             orig = container
             container = list(container)
-            while True:
-                try:
-                    container.remove(old)
-                except ValueError:
-                    break
+            for index in _find_sequence_indices(container, old):
+                del container[index]
             container = tuple(container)
             _swap_refs(orig, container, ignores)
         
-        elif containerType == dict:
+        elif containerType is dict or containerType is defaultdict:
             if "__internal_swaprefs_ignore__" not in container:
                 try:
-                    if old in container:
-                        container.pop(old)
+                    container.pop(old, None)
                 except TypeError:  # Unhashable old value
                     pass
                 for k,v in container.items():
                     if v is old:
                         del container[k]
 
-        elif containerType == set:
+        elif containerType is set:
             container.remove(old)
